import numpy as np
from time import time
from tqdm import tqdm
from keras.models import Model
from keras.layers import Input
from keras.optimizers import Adam
from keras.callbacks import TensorBoard
from keras import backend as K
from ISR.utils.datahandler import DataHandler
from ISR.utils.train_helper import TrainerHelper
from ISR.utils.metrics import PSNR
from ISR.utils.logger import get_logger


class Trainer:
    """Class object to setup and carry the training.

    Takes as input a generator that produces SR images.
    Conditionally, also a discriminator network and a feature extractor
        to build the components of the perceptual loss.
    Compiles the model(s) and trains in a GANS fashion if a discriminator is provided, otherwise
    carries a regular ISR training.

    Args:
        generator: Keras model, the super-scaling, or generator, network.
        discriminator: Keras model, the discriminator network for the adversarial
            component of the perceptual loss.
        feature_extractor: Keras model, feature extractor network for the deep features
            component of perceptual loss function.
        lr_train_dir: path to the directory containing the Low-Res images for training.
        hr_train_dir: path to the directory containing the High-Res images for training.
        lr_valid_dir: path to the directory containing the Low-Res images for validation.
        hr_valid_dir: path to the directory containing the High-Res images for validation.
        learning_rate: float.
        loss_weights: dictionary, use to weigh the components of the loss function.
            Contains 'MSE' for the MSE loss component, and can contain 'discriminator' and 'feat_extr'
            for the discriminator and deep features components respectively.
        logs_dir: path to the directory where the tensorboard logs are saved.
        weights_dir: path to the directory where the weights are saved.
        dataname: string, used to identify what dataset is used for the training session.
        weights_generator: path to the pre-trained generator's weights, for transfer learning.
        weights_discriminator: path to the pre-trained discriminator's weights, for transfer learning.
        n_validation:integer, number of validation samples used at training from the validation set.
        T: 0 < float <1, determines the 'flatness' threshold level for the training patches.
            See the TrainerHelper class for more details.
        lr_decay_frequency: integer, every how many epochs the learning rate is reduced.
        lr_decay_factor: 0 < float <1, learning rate reduction multiplicative factor.

    Methods:
        train: combines the networks and triggers training with the specified settings.

    """

    def __init__(
        self,
        generator,
        discriminator,
        feature_extractor,
        lr_train_dir,
        hr_train_dir,
        lr_valid_dir,
        hr_valid_dir,
        learning_rate=0.0004,
        loss_weights={'MSE': 1.0},
        logs_dir='logs',
        weights_dir='weights',
        dataname=None,
        weights_generator=None,
        weights_discriminator=None,
        n_validation=None,
        T=0.01,
        lr_decay_frequency=100,
        lr_decay_factor=0.5,
    ):
        if discriminator:
            assert generator.patch_size * generator.scale == discriminator.patch_size
        if feature_extractor:
            assert generator.patch_size * generator.scale == feature_extractor.patch_size
        self.generator = generator
        self.discriminator = discriminator
        self.feature_extractor = feature_extractor
        self.scale = generator.scale
        self.lr_patch_size = generator.patch_size
        self.learning_rate = learning_rate
        self.loss_weights = loss_weights
        self.best_metrics = {}
        self.pretrained_weights_path = {
            'generator': weights_generator,
            'discriminator': weights_discriminator,
        }
        self.lr_decay_factor = lr_decay_factor
        self.lr_decay_frequency = lr_decay_frequency

        self.helper = TrainerHelper(
            generator=self.generator,
            weights_dir=weights_dir,
            logs_dir=logs_dir,
            lr_train_dir=lr_train_dir,
            feature_extractor=self.feature_extractor,
            discriminator=self.discriminator,
            dataname=dataname,
            pretrained_weights_path=self.pretrained_weights_path,
        )

        self.model = self._combine_networks()

        self.train_dh = DataHandler(
            lr_dir=lr_train_dir,
            hr_dir=hr_train_dir,
            patch_size=self.lr_patch_size,
            scale=self.scale,
            n_validation_samples=None,
            T=T,
        )
        self.valid_dh = DataHandler(
            lr_dir=lr_valid_dir,
            hr_dir=hr_valid_dir,
            patch_size=self.lr_patch_size,
            scale=self.scale,
            n_validation_samples=n_validation,
            T=0.01,
        )
        self.logger = get_logger(__name__)

    def _combine_networks(self):
        """
        Constructs the combined model which contains the generator network,
        as well as discriminator and geature extractor, if any are defined.
        """

        lr = Input(shape=(self.lr_patch_size,) * 2 + (3,))
        sr = self.generator.model(lr)
        outputs = [sr]
        losses = ['mse']
        loss_weights = [self.loss_weights['MSE']]
        if self.discriminator:
            self.discriminator.model.trainable = False
            validity = self.discriminator.model(sr)
            outputs.append(validity)
            losses.append('binary_crossentropy')
            loss_weights.append(self.loss_weights['discriminator'])
        if self.feature_extractor:
            self.feature_extractor.model.trainable = False
            sr_feats = self.feature_extractor.model(sr)
            outputs.extend([*sr_feats])
            losses.extend(['mse'] * len(sr_feats))
            loss_weights.extend([self.loss_weights['feat_extr'] / len(sr_feats)] * len(sr_feats))
        combined = Model(inputs=lr, outputs=outputs)
        # https://stackoverflow.com/questions/42327543/adam-optimizer-goes-haywire-after-200k-batches-training-loss-grows
        optimizer = Adam(epsilon=0.0000001)
        combined.compile(
            loss=losses, loss_weights=loss_weights, optimizer=optimizer, metrics={'generator': PSNR}
        )
        return combined

    def _lr_scheduler(self, epoch):
        """ Scheduler for the learning rate updates. """

        n_decays = epoch // self.lr_decay_frequency
        # no lr below minimum control 10e-6
        return max(1e-6, self.learning_rate * (self.lr_decay_factor ** n_decays))

    def _load_weights(self):
        """
        Loads the pretrained weights from the given path, if any is provided.
        If a discriminator is defined, does the same.
        """

        gen_w = self.pretrained_weights_path['generator']
        if gen_w:
            self.model.get_layer('generator').load_weights(gen_w)
        if self.discriminator:
            dis_w = self.pretrained_weights_path['discriminator']
            if dis_w:
                self.model.get_layer('discriminator').load_weights(dis_w)
                self.discriminator.model.load_weights(dis_w)

    def train(self, epochs, steps_per_epoch, batch_size):
        """
        Carries on the training for the given number of epochs.
        Sends the losses to Tensorboard.
        """

        starting_epoch = self.helper.initialize_training(
            self
        )  # load_weights, creates folders, creates basename
        self.tensorboard = TensorBoard(log_dir=self.helper.callback_paths['logs'])
        self.tensorboard.set_model(self.model)

        # validation data
        validation_set = self.valid_dh.get_validation_set(batch_size)
        y_validation = [validation_set['hr']]
        if self.discriminator:
            discr_out_shape = list(self.discriminator.model.outputs[0].shape)[1:4]
            valid = np.ones([batch_size] + discr_out_shape)
            fake = np.zeros([batch_size] + discr_out_shape)
            validation_valid = np.ones([len(validation_set['hr'])] + discr_out_shape)
            y_validation.append(validation_valid)
        if self.feature_extractor:
            validation_feats = self.feature_extractor.model.predict(validation_set['hr'])
            y_validation.extend([*validation_feats])

        for epoch in range(starting_epoch, epochs):
            self.logger.info('Epoch {e}/{tot_eps}'.format(e=epoch, tot_eps=epochs))
            K.set_value(self.model.optimizer.lr, self._lr_scheduler(epoch=epoch))
            self.logger.info('Current learning rate: {}'.format(K.eval(self.model.optimizer.lr)))
            epoch_start = time()
            for step in tqdm(range(steps_per_epoch)):
                batch = self.train_dh.get_batch(batch_size)
                sr = self.generator.model.predict(batch['lr'])
                y_train = [batch['hr']]
                losses = {}

                ## Discriminator training
                if self.discriminator:
                    d_loss_real = self.discriminator.model.train_on_batch(batch['hr'], valid)
                    d_loss_fake = self.discriminator.model.train_on_batch(sr, fake)
                    d_loss_real = dict(
                        zip(
                            ['train_d_real_' + m for m in self.discriminator.model.metrics_names],
                            d_loss_real,
                        )
                    )
                    d_loss_fake = dict(
                        zip(
                            ['train_d_fake_' + m for m in self.discriminator.model.metrics_names],
                            d_loss_fake,
                        )
                    )
                    losses.update(d_loss_real)
                    losses.update(d_loss_fake)
                    y_train.append(valid)

                ## Generator training
                if self.feature_extractor:
                    hr_feats = self.feature_extractor.model.predict(batch['hr'])
                    y_train.extend([*hr_feats])

                trainig_loss = self.model.train_on_batch(batch['lr'], y_train)
                losses.update(
                    dict(zip(['train_' + m for m in self.model.metrics_names], trainig_loss))
                )
                self.tensorboard.on_epoch_end(epoch * steps_per_epoch + step, losses)
                self.logger.debug('Losses at step {s}:\n {l}'.format(s=step, l=losses))

            elapsed_time = time() - epoch_start
            self.logger.info('Epoch {} took {:10.1f}s'.format(epoch, elapsed_time))

            validation_loss = self.model.evaluate(
                validation_set['lr'], y_validation, batch_size=batch_size
            )
            losses = dict(zip(['val_' + m for m in self.model.metrics_names], validation_loss))

            monitored_metrics = {}
            if (not self.discriminator) and (not self.feature_extractor):
                monitored_metrics.update({'val_loss': 'min'})
            else:
                monitored_metrics.update({'val_generator_loss': 'min'})

            self.helper.on_epoch_end(
                epoch=epoch,
                losses=losses,
                generator=self.model.get_layer('generator'),
                discriminator=self.discriminator,
                metrics=monitored_metrics,
            )
            self.tensorboard.on_epoch_end(epoch, losses)
        self.tensorboard.on_train_end(None)



import numpy as np


def process_array(image_array):
    """ Process a 3-dimensional array into a scaled, 4 dimensional batch of size 1. """

    image_array = image_array / 255.0
    image_batch = np.expand_dims(image_array, axis=0)
    return image_batch


def process_output(output_tensor):
    """ Transforms the 4-dimensional output tensor into a suitable image format. """

    sr_img = output_tensor[0].clip(0, 1) * 255
    sr_img = np.uint8(sr_img)
    return sr_img



import tensorflow as tf
from keras.layers import UpSampling2D, concatenate, Input, Activation, Add, Conv2D, Lambda
from keras.models import Model
from ISR.models.imagemodel import ImageModel


def make_model(arch_params, patch_size):
    """ Returns the model.

    Used to select the model.
    """

    return RRDN(arch_params, patch_size)


class RRDN(ImageModel):
    """Implementation of the Residual in Residual Dense Network for image super-scaling.

    The network is the one described in https://arxiv.org/abs/1809.00219 (Wang et al. 2018).

    Args:
        arch_params: dictionary, contains the network parameters C, D, G, G0, T, x.
        patch_size: integer or None, determines the input size. Only needed at
            training time, for prediction is set to None.
        beta: float <= 1, scaling parameter for the residual connections.
        c_dim: integer, number of channels of the input image.
        kernel_size: integer, common kernel size for convolutions.
        upscaling: string, 'ups' or 'shuffle', determines which implementation
            of the upscaling layer to use.

    Attributes:
        C: integer, number of conv layer inside each residual dense blocks (RDB).
        D: integer, number of RDBs inside each Residual in Residual Dense Block (RRDB).
        T: integer, number or RRDBs.
        G: integer, number of convolution output filters inside the RDBs.
        G0: integer, number of output filters of each RDB.
        x: integer, the scaling factor.
        model: Keras model of the RRDN.
        name: name used to identify what upscaling network is used during training.
        model.name: identifies this network as the generator network
            in the compound model built by the trainer class.
    """

    def __init__(
        self, arch_params={}, patch_size=None, beta=0.2, c_dim=3, kernel_size=3
    ):  # , learning_rate=1e-5):
        self.params = arch_params
        self.beta = beta
        self.c_dim = c_dim
        self.C = self.params['C']
        self.D = self.params['D']
        self.G = self.params['G']
        self.G0 = self.params['G0']
        self.T = self.params['T']
        self.scale = self.params['x']
        self.kernel_size = kernel_size
        self.patch_size = patch_size
        self.model = self._build_rdn()
        self.model.name = 'generator'
        self.name = 'rrdn'

    def _dense_block(self, input_layer, d, t):
        """
        Implementation of the (Residual) Dense Block as in the paper
        Residual Dense Network for Image Super-Resolution (Zhang et al. 2018).

        Residuals are incorporated in the RRDB.
        d is an integer only used for naming. (d-th block)
        """

        x = input_layer
        for c in range(1, self.C + 1):
            F_dc = Conv2D(
                self.G, kernel_size=self.kernel_size, padding='same', name='F_%d_%d_%d' % (t, d, c)
            )(x)
            F_dc = Activation('relu', name='F_%d_%d_%d_Relu' % (t, d, c))(F_dc)
            x = concatenate([x, F_dc], axis=3, name='RDB_Concat_%d_%d_%d' % (t, d, c))

        # DIFFERENCE: in RDN a kernel size of 1 instead of 3 is used here
        x = Conv2D(self.G0, kernel_size=3, padding='same', name='LFF_%d_%d' % (t, d))(x)
        return x

    def _RRDB(self, input_layer, t):
        """Residual in Residual Dense Block.

        t is integer, for naming of RRDB.
        beta is scalar.
        """

        # SUGGESTION: MAKE BETA LEARNABLE
        x = input_layer

        for d in range(1, self.D + 1):
            LFF = self._dense_block(x, d, t)
            LFF_beta = Lambda(lambda x: x * self.beta)(LFF)
            x = Add(name='LRL_%d_%d' % (t, d))([x, LFF_beta])
        x = Lambda(lambda x: x * self.beta)(x)
        x = Add(name='RRDB_%d_out' % (t))([input_layer, x])
        return x

    def _pixel_shuffle(self, input_layer):
        """ PixelShuffle implementation of the upscaling part. """

        x = Conv2D(self.c_dim * self.scale ** 2, kernel_size=3, padding='same', name='PreShuffle')(
            input_layer
        )
        return Lambda(
            lambda x: tf.depth_to_space(x, block_size=self.scale, data_format='NHWC'),
            name='PixelShuffle',
        )(x)

    def _build_rdn(self):
        LR_input = Input(shape=(self.patch_size, self.patch_size, 3), name='LR_input')
        pre_blocks = Conv2D(
            self.G0, kernel_size=self.kernel_size, padding='same', name='Pre_blocks_conv'
        )(LR_input)
        # DIFFERENCE: in RDN an extra convolution is present here
        for t in range(1, self.T + 1):
            if t == 1:
                x = self._RRDB(pre_blocks, t)
            else:
                x = self._RRDB(x, t)
        # DIFFERENCE: in RDN a conv with kernel size of 1 after a concat operation is used here
        post_blocks = Conv2D(self.G0, kernel_size=3, padding='same', name='post_blocks_conv')(x)
        # Global Residual Learning
        GRL = Add(name='GRL')([post_blocks, pre_blocks])
        # Upscaling
        PS = self._pixel_shuffle(GRL)
        # Compose SR image
        SR = Conv2D(self.c_dim, kernel_size=self.kernel_size, padding='same', name='SR')(PS)
        return Model(inputs=LR_input, outputs=SR)



import os
import imageio
import numpy as np
from ISR.utils.logger import get_logger


class DataHandler:
    """
    DataHandler generate augmented batches used for training or validation.

    Args:
        lr_dir: directory containing the Low Res images.
        hr_dir: directory containing the High Res images.
        patch_size: integer, size of the patches extracted from LR images.
        scale: integer, upscaling factor.
        n_validation_samples: integer, size of the validation set. Only provided if the
            DataHandler is used to generate validation sets.
        T: float in [0,1], is the patch "flatness" threshold.
            Determines what level of detail the patches need to meet. 0 means any patch is accepted.
    """

    def __init__(self, lr_dir, hr_dir, patch_size, scale, n_validation_samples=None, T=0.03):
        self.folders = {'hr': hr_dir, 'lr': lr_dir}  # image folders
        self.extensions = ('.png', '.jpeg', '.jpg')  # admissible extension
        self.img_list = {}  # list of file names
        self.n_validation_samples = n_validation_samples
        self.patch_size = patch_size
        self.scale = scale
        self.T = T
        self.patch_size = {'lr': patch_size, 'hr': patch_size * self.scale}
        self.logger = get_logger(__name__)
        self._make_img_list()
        self._check_dataset()

    def _make_img_list(self):
        """ Creates a dictionary of lists of the acceptable images contained in lr_dir and hr_dir. """

        for res in ['hr', 'lr']:
            file_names = os.listdir(self.folders[res])
            file_names = [file for file in file_names if file.endswith(self.extensions)]
            self.img_list[res] = np.sort(file_names)

        if self.n_validation_samples:
            samples = np.random.choice(
                range(len(self.img_list['hr'])), self.n_validation_samples, replace=False
            )
            for res in ['hr', 'lr']:
                self.img_list[res] = self.img_list[res][samples]

    def _check_dataset(self):
        """ Sanity check for dataset. """

        # the order of these asserts is important for testing
        assert len(self.img_list['hr']) == self.img_list['hr'].shape[0], 'UnevenDatasets'
        assert self._matching_datasets(), 'Input/LabelsMismatch'

    def _matching_datasets(self):
        """ Rough file name matching between lr and hr directories. """
        # LR_name.png = HR_name+x+scale.png
        # or
        # LR_name.png = HR_name.png
        LR_name_root = [x.split('.')[0].split('x')[0] for x in self.img_list['lr']]
        HR_name_root = [x.split('.')[0] for x in self.img_list['hr']]
        return np.all(HR_name_root == LR_name_root)

    def _not_flat(self, patch):
        """
        Determines whether the patch is complex, or not-flat enough.
        Threshold set by self.T.
        """

        if max(np.std(patch, axis=0).mean(), np.std(patch, axis=1).mean()) < self.T:
            return False
        else:
            return True

    def _crop_imgs(self, imgs, batch_size, idx=0):
        """
        Get random top left corners coordinates in LR space, multiply by scale to
        get HR coordinates.
        Gets batch_size + n possible coordinates.
        Accepts the batch only if the standard deviation of pixel intensities is above a given threshold, OR
        no patches can be further discarded (n have been discarded already).
        Square crops of size patch_size are taken from the selected
        top left corners.
        """

        n = 2 * batch_size
        top_left = {'x': {}, 'y': {}}
        for i, axis in enumerate(['x', 'y']):
            top_left[axis]['lr'] = np.random.randint(
                0, imgs['lr'].shape[i] - self.patch_size['lr'] + 1, batch_size + n
            )
            top_left[axis]['hr'] = top_left[axis]['lr'] * self.scale

        crops = {}
        for res in ['lr', 'hr']:
            slices = [
                [slice(x, x + self.patch_size[res]), slice(y, y + self.patch_size[res])]
                for x, y in zip(top_left['x'][res], top_left['y'][res])
            ]
            crops[res] = []
            for s in slices:
                candidate_crop = imgs[res][s[0], s[1], slice(None)]
                if self._not_flat(candidate_crop) or n == 0:
                    crops[res].append(candidate_crop)
                else:
                    n -= 1
                if len(crops[res]) == batch_size:
                    break
            crops[res] = np.array(crops[res])
        return crops

    def _apply_transform(self, img, transform_selection):
        """ Rotates and flips input image according to transform_selection. """

        rotate = {
            0: lambda x: x,
            1: lambda x: np.rot90(x, k=1, axes=(1, 0)),  # rotate right
            2: lambda x: np.rot90(x, k=1, axes=(0, 1)),  # rotate left
        }

        flip = {
            0: lambda x: x,
            1: lambda x: np.flip(x, 0),  # flip along horizontal axis
            2: lambda x: np.flip(x, 1),  # flip along vertical axis
        }

        rot_direction = transform_selection[0]
        flip_axis = transform_selection[1]

        img = rotate[rot_direction](img)
        img = flip[flip_axis](img)

        return img

    def _transform_batch(self, batch, transforms):
        """ Transforms each individual image of the batch independently. """

        t_batch = np.array(
            [self._apply_transform(img, transforms[i]) for i, img in enumerate(batch)]
        )
        return t_batch

    def get_batch(self, batch_size, idx=None):
        """
        Returns a dictionary with keys ('lr', 'hr') containing training batches
        of Low Res and High Res image patches.
        """

        if not idx:
            # randomly select one image. idx is given at validation time.
            idx = np.random.choice(range(len(self.img_list['hr'])))
        img = {}
        for res in ['lr', 'hr']:
            img_path = os.path.join(self.folders[res], self.img_list[res][idx])
            img[res] = imageio.imread(img_path) / 255.0
        batch = self._crop_imgs(img, batch_size)
        transforms = np.random.randint(0, 3, (batch_size, 2))
        batch['lr'] = self._transform_batch(batch['lr'], transforms)
        batch['hr'] = self._transform_batch(batch['hr'], transforms)

        return batch

    def get_validation_batches(self, batch_size):
        """ Returns a batch for each image in the validation set. """

        if self.n_validation_samples:
            batches = []
            for idx in range(self.n_validation_samples):
                batches.append(self.get_batch(batch_size, idx))
            return batches
        else:
            self.logger.error(
                'No validation set size specified. (not operating in a validation set?)'
            )
            raise ValueError(
                'No validation set size specified. (not operating in a validation set?)'
            )

    def get_validation_set(self, batch_size):
        """
        Returns a batch for each image in the validation set.
        Flattens and splits them to feed it to Keras's model.evaluate.
        """

        if self.n_validation_samples:
            batches = self.get_validation_batches(batch_size)
            valid_set = {'lr': [], 'hr': []}
            for batch in batches:
                for res in ('lr', 'hr'):
                    valid_set[res].extend(batch[res])
            for res in ('lr', 'hr'):
                valid_set[res] = np.array(valid_set[res])
            return valid_set
        else:
            self.logger.error(
                'No validation set size specified. (not operating in a validation set?)'
            )
            raise ValueError(
                'No validation set size specified. (not operating in a validation set?)'
            )



import tensorflow as tf
from keras.layers import concatenate, Input, Activation, Add, Conv2D, Lambda, UpSampling2D
from keras.models import Model
from ISR.models.imagemodel import ImageModel


def make_model(arch_params, patch_size):
    """ Returns the model.

    Used to select the model.
    """

    return RDN(arch_params, patch_size)


class RDN(ImageModel):
    """Implementation of the Residual Dense Network for image super-scaling.

    The network is the one described in https://arxiv.org/abs/1802.08797 (Zhang et al. 2018).

    Args:
        arch_params: dictionary, contains the network parameters C, D, G, G0, x.
        patch_size: integer or None, determines the input size. Only needed at
            training time, for prediction is set to None.
        c_dim: integer, number of channels of the input image.
        kernel_size: integer, common kernel size for convolutions.
        upscaling: string, 'ups' or 'shuffle', determines which implementation
            of the upscaling layer to use.

    Attributes:
        C: integer, number of conv layer inside each residual dense blocks (RDB).
        D: integer, number of RDBs.
        G: integer, number of convolution output filters inside the RDBs.
        G0: integer, number of output filters of each RDB.
        x: integer, the scaling factor.
        model: Keras model of the RDN.
        name: name used to identify what upscaling network is used during training.
        model.name: identifies this network as the generator network
            in the compound model built by the trainer class.
    """

    def __init__(self, arch_params={}, patch_size=None, c_dim=3, kernel_size=3, upscaling='ups'):
        self.params = arch_params
        self.C = self.params['C']
        self.D = self.params['D']
        self.G = self.params['G']
        self.G0 = self.params['G0']
        self.scale = self.params['x']
        self.patch_size = patch_size
        self.c_dim = c_dim
        self.kernel_size = kernel_size
        self.upscaling = upscaling
        self.model = self._build_rdn()
        self.model.name = 'generator'
        self.name = 'rdn'

    def _upsampling_block(self, input_layer):
        """ Upsampling block for old weights. """

        x = Conv2D(self.c_dim * self.scale ** 2, kernel_size=3, padding='same', name='UPN3')(
            input_layer
        )
        return UpSampling2D(size=self.scale, name='UPsample')(x)

    def _pixel_shuffle(self, input_layer):
        """ PixelShuffle implementation of the upscaling layer. """

        x = Conv2D(self.c_dim * self.scale ** 2, kernel_size=3, padding='same', name='UPN3')(
            input_layer
        )
        return Lambda(
            lambda x: tf.depth_to_space(x, block_size=self.scale, data_format='NHWC'),
            name='PixelShuffle',
        )(x)

    def _UPN(self, input_layer):
        """ Upscaling layers. With old weights use _upsampling_block instead of _pixel_shuffle. """

        x = Conv2D(64, kernel_size=5, strides=1, padding='same', name='UPN1')(input_layer)
        x = Activation('relu', name='UPN1_Relu')(x)
        x = Conv2D(32, kernel_size=3, padding='same', name='UPN2')(x)
        x = Activation('relu', name='UPN2_Relu')(x)
        if self.upscaling == 'shuffle':
            return self._pixel_shuffle(x)
        elif self.upscaling == 'ups':
            return self._upsampling_block(x)
        else:
            raise ValueError('Invalid choice of upscaling layer.')

    def _RDBs(self, input_layer):
        """RDBs blocks.

        Args:
            input_layer: input layer to the RDB blocks (e.g. the second convolutional layer F_0).

        Returns:
            concatenation of RDBs output feature maps with G0 feature maps.
        """
        rdb_concat = list()
        rdb_in = input_layer
        for d in range(1, self.D + 1):
            x = rdb_in
            for c in range(1, self.C + 1):
                F_dc = Conv2D(
                    self.G, kernel_size=self.kernel_size, padding='same', name='F_%d_%d' % (d, c)
                )(x)
                F_dc = Activation('relu', name='F_%d_%d_Relu' % (d, c))(F_dc)
                # concatenate input and output of ConvRelu block
                # x = [input_layer,F_11(input_layer),F_12([input_layer,F_11(input_layer)]), F_13..]
                x = concatenate([x, F_dc], axis=3, name='RDB_Concat_%d_%d' % (d, c))
            # 1x1 convolution (Local Feature Fusion)
            x = Conv2D(self.G0, kernel_size=1, name='LFF_%d' % (d))(x)
            # Local Residual Learning F_{i,LF} + F_{i-1}
            rdb_in = Add(name='LRL_%d' % (d))([x, rdb_in])
            rdb_concat.append(rdb_in)

        assert len(rdb_concat) == self.D

        return concatenate(rdb_concat, axis=3, name='LRLs_Concat')

    def _build_rdn(self):
        LR_input = Input(shape=(self.patch_size, self.patch_size, 3), name='LR')
        F_m1 = Conv2D(self.G0, kernel_size=self.kernel_size, padding='same', name='F_m1')(LR_input)
        F_0 = Conv2D(self.G0, kernel_size=self.kernel_size, padding='same', name='F_0')(F_m1)
        FD = self._RDBs(F_0)
        # Global Feature Fusion
        # 1x1 Conv of concat RDB layers -> G0 feature maps
        GFF1 = Conv2D(self.G0, kernel_size=1, padding='same', name='GFF_1')(FD)
        GFF2 = Conv2D(self.G0, kernel_size=self.kernel_size, padding='same', name='GFF_2')(GFF1)
        # Global Residual Learning for Dense Features
        FDF = Add(name='FDF')([GFF2, F_m1])
        # Upscaling
        FU = self._UPN(FDF)
        # Compose SR image
        SR = Conv2D(self.c_dim, kernel_size=self.kernel_size, padding='same', name='SR')(FU)

        return Model(inputs=LR_input, outputs=SR)



from keras.layers import concatenate, Flatten, Input, Activation, Dense, Conv2D, BatchNormalization
from keras.layers.advanced_activations import LeakyReLU
from keras.models import Model
from keras.optimizers import Adam


class Discriminator:
    """
    Implementation of the discriminator network for the adversarial
    component of the perceptual loss.

    Args:
        patch_size: integer, determines input size as (patch_size, patch_size, 3).
        kernel_size: size of the kernel in the conv blocks.

    Attributes:
        model: Keras model.
        name: name used to identify what discriminator is used during GANs training.
        model.name: identifies this network as the discriminator network
            in the compound model built by the trainer class.
        block_param: dictionary, determines the number of filters and the strides for each
            conv block.

    """

    def __init__(self, patch_size, kernel_size=3):
        self.patch_size = patch_size
        self.kernel_size = kernel_size
        self.block_param = {}
        self.block_param['filters'] = (64, 128, 128, 256, 256, 512, 512)
        self.block_param['strides'] = (2, 1, 2, 1, 1, 1, 1)
        self.block_num = len(self.block_param['filters'])
        self.model = self._build_disciminator()
        optimizer = Adam(0.0002, 0.5)
        self.model.compile(loss='binary_crossentropy', optimizer=optimizer, metrics=['accuracy'])
        self.model.name = 'discriminator'
        self.name = 'srgan-large'

    def _conv_block(self, input, filters, strides, batch_norm=True, count=None):
        """ Convolutional layer + Leaky ReLU + conditional BN. """

        x = Conv2D(
            filters,
            kernel_size=self.kernel_size,
            strides=strides,
            padding='same',
            name='Conv_{}'.format(count),
        )(input)
        x = LeakyReLU(alpha=0.2)(x)
        if batch_norm:
            x = BatchNormalization(momentum=0.8)(x)
        return x

    def _build_disciminator(self):
        """ Puts the discriminator's layers together. """

        HR = Input(shape=(self.patch_size, self.patch_size, 3))
        x = self._conv_block(HR, filters=64, strides=1, batch_norm=False, count=1)
        for i in range(self.block_num):
            x = self._conv_block(
                x,
                filters=self.block_param['filters'][i],
                strides=self.block_param['strides'][i],
                count=i + 2,
            )
        x = Dense(self.block_param['filters'][-1] * 2, name='Dense_1024')(x)
        x = LeakyReLU(alpha=0.2)(x)
        # x = Flatten()(x)
        x = Dense(1, name='Dense_last')(x)
        HR_v_SR = Activation('sigmoid')(x)

        discriminator = Model(inputs=HR, outputs=HR_v_SR)
        return discriminator



import os
import argparse
import numpy as np
import yaml
from ISR.utils.logger import get_logger


logger = get_logger(__name__)


def _get_parser():
    parser = argparse.ArgumentParser()
    parser.add_argument('--prediction', action='store_true', dest='prediction')
    parser.add_argument('--training', action='store_true', dest='training')
    parser.add_argument('--summary', action='store_true', dest='summary')
    parser.add_argument('--default', action='store_true', dest='default')
    parser.add_argument('--config', action='store', dest='config_file')
    return parser


def parse_args():
    """ Parse CLI arguments. """
    parser = _get_parser()
    args = vars(parser.parse_args())
    if args['prediction'] and args['training']:
        logger.error('Select only prediction OR training.')
        raise ValueError('Select only prediction OR training.')
    return args


def get_config_from_weights(w_path, arch_params, name):
    """
    Extracts architecture parameters from the file name of the weights.
    Only works with standardized weights name.
    """

    w_path = os.path.basename(w_path)
    parts = w_path.split(name)[1]
    parts = parts.split('_')[0]
    parts = parts.split('-')
    new_param = {}
    for param in arch_params:
        param_part = [x for x in parts if param in x]
        param_value = int(param_part[0].split(param)[1])
        new_param[param] = param_value
    return new_param


def select_option(options, message='', val=None):
    """ CLI selection given options. """

    while val not in options:
        val = input(message)
        if val not in options:
            logger.error('Invalid choice.')
    return val


def select_positive(message='', value=-1):
    """ CLI non-negative integer selection. """

    while value < 0:
        value = int(input(message))
        if value < 0:
            logger.error('Invalid choice.')
    return value


def browse_weights(weights_dir, model='generator'):
    """ Weights selection from cl. """

    exit = False
    while exit is False:
        weights = np.sort(os.listdir(weights_dir))[::-1]
        print_sel = dict(zip(np.arange(len(weights)), weights))
        for k in print_sel.keys():
            logger_message = '{item_n}: {item} \n'.format(item_n=k, item=print_sel[k])
            logger.info(logger_message)

        sel = select_positive('>>> Select folder or weights for {}\n'.format(model))
        if weights[sel].endswith('hdf5'):
            weights_path = os.path.join(weights_dir, weights[sel])
            exit = True
        else:
            weights_dir = os.path.join(weights_dir, weights[sel])
    return weights_path


def setup(config_file='config.yml', default=False, training=False, prediction=False):
    """CLI interface to set up the training or prediction session.

    Takes as input the configuration file path (minus the '.py' extension)
    and arguments parse from CLI.
    """

    conf = yaml.load(open(config_file, 'r'), Loader=yaml.FullLoader)
    if training:
        session_type = 'training'
    elif prediction:
        session_type = 'prediction'
    else:
        message = '(t)raining or (p)rediction? (t/p) '
        session_type = {'t': 'training', 'p': 'prediction'}[select_option(['t', 'p'], message)]
    if default:
        all_default = 'y'
    else:
        all_default = input('Default options for everything? (y/[n]) ')

    if all_default == 'y':
        generator = conf['default']['generator']
        if session_type == 'prediction':
            dataset = conf['default']['test_set']
            conf['generators'][generator] = get_config_from_weights(
                conf['weights_paths']['generator'], conf['generators'][generator], generator
            )
        elif session_type == 'training':
            dataset = conf['default']['training_set']

        return session_type, generator, conf, dataset

    logger.info('Select SR (generator) network')
    generators = {}
    for i, gen in enumerate(conf['generators']):
        generators[str(i)] = gen
        logger.info('{}: {}'.format(i, gen))
    generator = generators[select_option(generators)]

    load_weights = input('Load pretrained weights for {}? ([y]/n/d) '.format(generator))
    if load_weights == 'n':
        default = input('Load default parameters for {}? ([y]/n) '.format(generator))
        if default == 'n':
            for param in conf['generators'][generator]:
                value = select_positive(message='{}:'.format(param))
                conf['generators'][generator][param] = value
        else:
            logger.info('Default {} parameters.'.format(generator))
    elif (load_weights == 'd') and (conf['weights_paths']['generator']):
        logger.info('Loading default weights for {}'.format(generator))
        logger.info(conf['weights_paths']['generator'])
        conf['generators'][generator] = get_config_from_weights(
            conf['weights_paths']['generator'], conf['generators'][generator], generator
        )
    else:
        conf['weights_paths']['generator'] = browse_weights(conf['dirs']['weights'], generator)
        conf['generators']['generator'] = get_config_from_weights(
            conf['weights_paths']['generator'], conf['generators'][generator], generator
        )
    logger.info('{} parameters:'.format(generator))
    logger.info(conf['generators'][generator])
    if session_type == 'training':
        use_discr = input('Use an Adversarial Network? (y/[n]) ')
        if use_discr == 'y':
            conf['default']['discriminator'] = True
            discr_w = input('Use pretrained discriminator weights? (y/[n]) ')
            if discr_w == 'y':
                conf['weights_paths']['discriminator'] = browse_weights(
                    conf['dirs']['weights'], 'discriminator'
                )

        use_feat_ext = input('Use feature extractor? (y/[n]) ')
        if use_feat_ext == 'y':
            conf['default']['feat_ext'] = True
    dataset = select_dataset(session_type, conf)

    return session_type, generator, conf, dataset


def select_dataset(session_type, conf):
    """ CLI snippet for selection the dataset for training. """

    if session_type == 'training':
        logger.info('Select training set')
        datasets = {}
        for i, data in enumerate(conf['training_sets']):
            datasets[str(i)] = data
            logger.info('{}: {}'.format(i, data))
        dataset = datasets[select_option(datasets)]

        return dataset
    else:
        logger.info('Select test set')
        datasets = {}
        for i, data in enumerate(conf['test_sets']):
            datasets[str(i)] = data
            logger.info('{}: {}'.format(i, data))
        dataset = datasets[select_option(datasets)]

        return dataset



import imageio
import os
import numpy as np
from time import time
from ISR.utils.logger import get_logger
from ISR.utils.image_processing import process_array, process_output


class Predictor:
    """The predictor class handles prediction, given an input model.

    Loads the images in the input directory, executes training given a model
    and saves the results in the output directory.
    Can receive a path for the weights or can let the user browse through the
    weights directory for the desired weights.

    Args:
        input_dir: string, path to the input directory.
        output_dir: string, path to the output directory.
        verbose: bool.

    Attributes:
        extensions: list of accepted image extensions.
        img_ls: list of image files in input_dir.

    Methods:
        get_predictions: given a model and a string containing the weights' path,
            runs the predictions on the images contained in the input directory and
            stores the results in the output directory.
    """

    def __init__(self, input_dir, output_dir='./data/output', verbose=True):

        self.input_dir = input_dir
        self.data_name = os.path.basename(os.path.normpath(self.input_dir))
        self.output_dir = os.path.join(output_dir, self.data_name)
        file_ls = os.listdir(self.input_dir)
        self.logger = get_logger(__name__)
        if not verbose:
            self.logger.setLevel(40)
        self.extensions = ('.jpeg', '.jpg', '.png')  # file extensions that are admitted
        self.img_ls = [file for file in file_ls if file.endswith(self.extensions)]
        if len(self.img_ls) < 1:
            self.logger.error('No valid image files found (check config file).')
            raise ValueError('No valid image files found (check config file).')
        # Create results folder
        if not os.path.exists(self.output_dir):
            self.logger.info('Creating output directory:\n{}'.format(self.output_dir))
            os.makedirs(self.output_dir, exist_ok=True)

    def _load_weights(self):
        """ Invokes the model's load weights function if any weights are provided. """

        if self.weights_path is not None:
            self.logger.info('Loaded weights from \n > {}'.format(self.weights_path))
            # loading by name automatically excludes the vgg layers
            self.model.model.load_weights(self.weights_path)
        else:
            self.logger.error('Error: Weights path not specified (check config file).')
            raise ValueError('Weights path not specified (check config file).')

    def _make_directory_structure(self):
        """ Creates the folder structure from the weights' name. """

        filename = os.path.basename(self.weights_path)
        weights_name, _ = os.path.splitext(filename)
        subdirs = weights_name.split('_')
        self.basepath = os.path.join(*subdirs)

    def get_predictions(self, model, weights_path):
        """ Runs the prediction. """

        self.model = model
        self.weights_path = weights_path
        self._load_weights()
        self._make_directory_structure()
        out_folder = os.path.join(self.output_dir, self.basepath)
        self.logger.info('Results in:\n > {}'.format(out_folder))
        if os.path.exists(out_folder):
            self.logger.warning('Directory exists, might overwrite files')
        os.makedirs(out_folder, exist_ok=True)
        # Predict and store
        for img_name in self.img_ls:
            input_path = os.path.join(self.input_dir, img_name)
            output_path = os.path.join(out_folder, img_name)
            self.logger.info('Processing file\n > {}'.format(input_path))
            start = time()
            sr_img = self._forward_pass(input_path)
            end = time()
            self.logger.info('Elapsed time: {}s'.format(end - start))
            self.logger.info('Result in: {}'.format(output_path))
            imageio.imwrite(output_path, sr_img)

    def _forward_pass(self, file_path):
        lr_img = imageio.imread(file_path)
        if lr_img.shape[2] == 3:
            lr_img = process_array(lr_img)
            sr_img = self.model.model.predict(lr_img)
            sr_img = process_output(sr_img)
            return sr_img
        else:
            self.logger.error('{} is not an image with 3 channels.'.format(file_path))



from keras.models import Model
from keras.applications.vgg19 import VGG19
from keras.layers import Input
from ISR.utils.logger import get_logger


class Cut_VGG19:
    """
    Class object that fetches keras' VGG19 model trained on the imagenet dataset
    and declares <layers_to_extract> as output layers. Used as feature extractor
    for the perceptual loss function.

    Args:
        layers_to_extract: list of layers to be declared as output layers.
        patch_size: integer, defines the size of the input (patch_size x patch_size).

    Attributes:
        loss_model: multi-output vgg architecture with <layers_to_extract> as output layers.
    """

    def __init__(self, patch_size, layers_to_extract):
        self.patch_size = patch_size
        self.input_shape = (patch_size,) * 2 + (3,)
        self.layers_to_extract = layers_to_extract
        self.logger = get_logger(__name__)

        if len(self.layers_to_extract) > 0:
            self._cut_vgg()
        else:
            self.logger.error('Invalid VGG instantiation: extracted layer must be > 0')
            raise ValueError('Invalid VGG instantiation: extracted layer must be > 0')

    def _cut_vgg(self):
        """
        Loads pre-trained VGG, declares as output the intermediate
        layers selected by self.layers_to_extract.
        """

        vgg = VGG19(weights='imagenet', include_top=False)
        vgg.trainable = False
        collect = []
        for i in self.layers_to_extract:
            collect.append(vgg.layers[i].output)

        vgg.outputs = collect
        hr = Input(shape=self.input_shape)
        features = vgg(hr)
        self.model = Model(inputs=hr, outputs=features)
        self.model.name = 'feat_extr'
        self.name = 'vgg19'  # used in weights naming



import keras.backend as K


def PSNR(y_true, y_pred, MAXp=1):
    """
    Evaluates the PSNR value:
        PSNR = 20 * log10(MAXp) - 10 * log10(MSE).

    Args:
        y_true: ground truth.
        y_pred: predicted value.
        MAXp: maximum value of the pixel range (default=1).
    """
    return -10.0 * K.log(K.mean(K.square(y_pred - y_true))) / K.log(10.0)



from ISR.utils.image_processing import process_array, process_output


class ImageModel:
    """ISR models parent class.

    Contains functions that are common across the super-scaling models.
    """

    def predict(self, input_image_array):
        """
        Processes the image array into a suitable format
        and transforms the network output in a suitable image format.

        Args:
            input_image_array: input image array.
        Returns:
            sr_img: image output.
        """
        lr_img = process_array(input_image_array)
        sr_img = self.model.predict(lr_img)
        sr_img = process_output(sr_img)
        return sr_img



import os
import numpy as np
from ISR.utils.logger import get_logger


class TrainerHelper:
    """Collection of useful functions to manage training sessions.

    Args:
        generator: Keras model, the super-scaling, or generator, network.
        logs_dir: path to the directory where the tensorboard logs are saved.
        weights_dir: path to the directory where the weights are saved.
        lr_train_dir: path to the directory containing the Low-Res images.
        feature_extractor: Keras model, feature extractor network for the deep features
            component of perceptual loss function.
        discriminator: Keras model, the discriminator network for the adversarial
            component of the perceptual loss.
        dataname: string, used to identify what dataset is used for the training session.
        pretrained_weights_path: dictionary, contains the paths, if any to the
            pre-trained generator's and to the pre-trained discriminator's weights,
            for transfer learning.
        fallback_save_every_n_epochs: integer, determines after how many epochs that did not trigger
            weights saving the weights are despite no metric improvement.


    Methods:
        print_training_setting: see docstring.
        on_epoch_end: see docstring.
        epoch_n_from_weights_name: see docstring.
        initialize_training: see docstring.

    """

    def __init__(
        self,
        generator,
        weights_dir,
        logs_dir,
        lr_train_dir,
        feature_extractor=None,
        discriminator=None,
        dataname=None,
        pretrained_weights_path={},
        fallback_save_every_n_epochs=2,
    ):
        self.generator = generator
        self.dirs = {'logs': logs_dir, 'weights': weights_dir}
        self.feature_extractor = feature_extractor
        self.discriminator = discriminator
        self.dataname = dataname
        self.pretrained_weights_path = pretrained_weights_path
        self.lr_dir = lr_train_dir
        self.best_metrics = {}
        self.fallback_save_every_n_epochs = fallback_save_every_n_epochs
        self.since_last_epoch = 0
        self.logger = get_logger(__name__)

    def _make_generator_name(self):
        """ Combines generators's name and its architecture's parameters. """

        gen_name = self.generator.name
        params = [gen_name]
        for param in np.sort(list(self.generator.params.keys())):
            params.append('{g}{p}'.format(g=param, p=self.generator.params[param]))
        return '-'.join(params)

    def _make_basename(self):
        """
        Constructs the name of the training session from training settings
        and the pre-trained weights, if any.
        """

        current_training_setup = self.dataname
        if self.feature_extractor:
            params = [self.feature_extractor.name]
            for feat in self.feature_extractor.layers_to_extract:
                params.append(str(feat))
            feat_name = '-'.join(params)
            current_training_setup = '-'.join([current_training_setup, feat_name])
        if self.discriminator:
            current_training_setup = '-'.join([current_training_setup, self.discriminator.name])

        # append to old description if any, otherwise append to gen architecture (which is constant across trainings)
        if ('generator' in self.pretrained_weights_path) and self.pretrained_weights_path[
            'generator'
        ]:
            root = os.path.splitext(os.path.basename(self.pretrained_weights_path['generator']))[0]
        else:
            root = self.generator_description
        return '_'.join([root, current_training_setup])

    def _get_basepath(self):
        """ Creates the dir structure needed to save weights and logs from the basename. """

        subdirs = self.basename.split('_')
        return os.path.join(*subdirs)

    def _create_dirs(self):
        """ Creates the directories associated with the class instance. """

        for dir in self.dirs:
            if not os.path.exists(os.path.join(self.dirs[dir], self.basepath)):
                os.makedirs(os.path.join(self.dirs[dir], self.basepath), exist_ok=False)

    def _make_callback_paths(self):
        """ Creates the paths used for managing logs and weights storage. """

        callback_paths = {}
        callback_paths['weights'] = os.path.join(self.dirs['weights'], self.basepath)
        callback_paths['logs'] = os.path.join(self.dirs['logs'], self.basepath)
        return callback_paths

    def _weights_name(self):
        """ Builds the string used to name the weights of the training session. """

        w_name = {'generator': '-'.join([self.basename, 'e{epoch:03d}.hdf5'])}
        if self.discriminator:
            w_name.update(
                {'discriminator': '-'.join(['discr', self.basename, 'e{epoch:03d}.hdf5'])}
            )
        return w_name

    def print_training_setting(self):
        """ Does what it says. """

        self.logger.info('\nTraining details:')
        self.logger.info('Generator settings:\n{}'.format(self.generator.params))
        if self.pretrained_weights_path['generator']:
            self.logger.info('Generator pretrained weights:')
            self.logger.info(
                '>> {}'.format(os.path.basename(self.pretrained_weights_path['generator']))
            )
        if self.discriminator:
            self.logger.info('Using GAN discriminator.')
            if self.pretrained_weights_path['discriminator']:
                self.logger.info('Discriminator pretrained weights:')
                self.logger.info(
                    '>> {}'.format(os.path.basename(self.pretrained_weights_path['discriminator']))
                )
        if self.feature_extractor:
            self.logger.info('Using high level features loss:')
            self.logger.info(
                '{f_ext} layers: {layers}'.format(
                    f_ext=self.feature_extractor.model.name,
                    layers=self.feature_extractor.layers_to_extract,
                )
            )
        self.logger.info('Training session name identifier: {}'.format(self.basename))
        self.logger.info('Input dir: {}'.format(self.lr_dir))
        self.logger.info('Patch size: {}'.format(self.generator.patch_size))
        self.logger.info('Saving weights under: {}'.format(self.callback_paths['weights']))
        self.logger.info('Saving tensorboard logs under: {}'.format(self.callback_paths['logs']))

    def _save_weights(self, epoch, generator, discriminator=None, best=False):
        """ Saves the weights of the non-None models. """

        if best:
            w_name = 'BEST-' + self.weights_name['generator'].format(epoch=epoch + 1)
        else:
            w_name = self.weights_name['generator'].format(epoch=epoch + 1)
        gen_path = os.path.join(self.callback_paths['weights'], w_name)
        generator.save_weights(
            gen_path
        )  # CANT SAVE MODEL DUE TO TF LAYER INSIDE LAMBDA (PIXELSHUFFLE)
        if discriminator:
            if best:
                w_name = 'BEST-' + self.weights_name['discriminator'].format(epoch=epoch + 1)
            else:
                w_name = self.weights_name['discriminator'].format(epoch=epoch + 1)
            discr_path = os.path.join(self.callback_paths['weights'], w_name)
            discriminator.model.save_weights(discr_path)
        try:
            self._remove_old_weights(5)
        except Exception as e:
            self.logger.warning('Could not remove weights'.format(e))

    def _remove_old_weights(self, max_n_weights, max_best=3):
        """Scans the weights folder and removes all but:
            - the max_best newest 'best' weights.
            - max_n_weights most recent 'others' weights.
        """

        w_list = {}
        w_list['all'] = [
            w for w in os.listdir(self.callback_paths['weights']) if w.endswith('.hdf5')
        ]
        w_list['best'] = [w for w in w_list['all'] if w.startswith('BEST')]
        w_list['others'] = [w for w in w_list['all'] if w not in w_list['best']]
        # remove older best
        epochs_set = {}
        epochs_set['best'] = list(set([self.epoch_n_from_weights_name(w) for w in w_list['best']]))
        epochs_set['others'] = list(
            set([self.epoch_n_from_weights_name(w) for w in w_list['others']])
        )
        keep_max = {'best': max_best, 'others': max_n_weights}
        for type in ['others', 'best']:
            if len(epochs_set[type]) > keep_max[type]:
                epoch_list = np.sort(epochs_set[type])[::-1]
                epoch_list = epoch_list[0 : keep_max[type]]
                for w in w_list[type]:
                    if self.epoch_n_from_weights_name(w) not in epoch_list:
                        # print('remove', os.path.join(self.callback_paths['weights'], w))
                        os.remove(os.path.join(self.callback_paths['weights'], w))

    def on_epoch_end(
        self, epoch, losses, generator, discriminator=None, metrics={'val_generator_loss': 'min'}
    ):
        """
        Manages the operations that are taken at the end of each epoch:
        metric checks, weight saves, logging.
        """

        self.logger.info(losses)
        monitor_op = {'max': np.greater, 'min': np.less}
        extreme = {'max': -np.Inf, 'min': np.Inf}
        for metric in metrics:
            if metric in losses.keys():
                if metric not in self.best_metrics.keys():
                    self.best_metrics[metric] = extreme[metrics[metric]]

                if monitor_op[metrics[metric]](losses[metric], self.best_metrics[metric]):
                    self.logger.info(
                        '{} improved from {:10.5f} to {:10.5f}'.format(
                            metric, self.best_metrics[metric], losses[metric]
                        )
                    )
                    self.logger.info('Saving weights')
                    self.best_metrics[metric] = losses[metric]
                    self._save_weights(epoch, generator, discriminator, best=True)
                    self.since_last_epoch = 0
                    return True
                else:
                    self.logger.info('{} did not improve.'.format(metric))
                    if self.since_last_epoch >= self.fallback_save_every_n_epochs:
                        self.logger.info('Saving weights anyways.')
                        self._save_weights(epoch, generator, discriminator, best=False)
                        self.since_last_epoch = 0
                        return True

            else:
                self.logger.warning('{} is not monitored, cannot save weights.'.format(metric))
        self.since_last_epoch += 1
        return False

    def epoch_n_from_weights_name(self, w_name):
        """
        Extracts the last epoch number from the standardized weights name.
        Only works with standardized weights names.
        """

        return int(os.path.splitext(os.path.basename(w_name))[0].split('-')[-1][1:])

    def initialize_training(self, object):
        """Function that is exectured prior to training.

        Wraps up most of the functions of this class:
        load the weights if any are given, generaters names for session and weights,
        creates directories and prints the training session.
        """
        object.pretrained_weights_path = self.pretrained_weights_path
        object._load_weights()
        w_name = object.pretrained_weights_path['generator']
        if w_name:
            last_epoch = self.epoch_n_from_weights_name(w_name)
        else:
            last_epoch = 0

        self.generator_description = self._make_generator_name()
        self.basename = self._make_basename()
        self.basepath = self._get_basepath()
        self.weights_name = self._weights_name()
        self.callback_paths = self._make_callback_paths()
        self._create_dirs()
        self.print_training_setting()
        return last_epoch



import os
import numpy as np
from importlib import import_module
from ISR.utils.utils import setup, parse_args
from ISR.utils.logger import get_logger


def _get_module(generator):
    return import_module('ISR.models.' + generator)


def run(config_file, default=False, training=False, prediction=False):
    os.environ['TF_CPP_MIN_LOG_LEVEL'] = '3'
    logger = get_logger(__name__)
    session_type, generator, conf, dataset = setup(config_file, default, training, prediction)

    lr_patch_size = conf['session'][session_type]['patch_size']
    scale = conf['generators'][generator]['x']

    module = _get_module(generator)
    gen = module.make_model(conf['generators'][generator], lr_patch_size)
    if session_type == 'prediction':
        from ISR.predict.predictor import Predictor

        pr_h = Predictor(input_dir=conf['test_sets'][dataset])
        pr_h.get_predictions(gen, conf['weights_paths']['generator'])

    elif session_type == 'training':
        from ISR.train.trainer import Trainer

        hr_patch_size = lr_patch_size * scale
        if conf['default']['feat_ext']:
            from ISR.models.cut_vgg19 import Cut_VGG19

            out_layers = conf['feat_extr']['vgg19']['layers_to_extract']
            f_ext = Cut_VGG19(patch_size=hr_patch_size, layers_to_extract=out_layers)
        else:
            f_ext = None

        if conf['default']['discriminator']:
            from ISR.models.discriminator import Discriminator

            discr = Discriminator(patch_size=hr_patch_size, kernel_size=3)
        else:
            discr = None

        trainer = Trainer(
            generator=gen,
            discriminator=discr,
            feature_extractor=f_ext,
            lr_train_dir=conf['training_sets'][dataset]['lr_train_dir'],
            hr_train_dir=conf['training_sets'][dataset]['hr_train_dir'],
            lr_valid_dir=conf['training_sets'][dataset]['lr_valid_dir'],
            hr_valid_dir=conf['training_sets'][dataset]['hr_valid_dir'],
            loss_weights=conf['loss_weights'],
            dataname=conf['training_sets'][dataset]['data_name'],
            logs_dir=conf['dirs']['logs'],
            weights_dir=conf['dirs']['weights'],
            weights_generator=conf['weights_paths']['generator'],
            weights_discriminator=conf['weights_paths']['discriminator'],
            n_validation=conf['session'][session_type]['n_validation_samples'],
            lr_decay_frequency=conf['session'][session_type]['lr_decay_frequency'],
            lr_decay_factor=conf['session'][session_type]['lr_decay_factor'],
            T=0.01,
        )
        trainer.train(
            epochs=conf['session'][session_type]['epochs'],
            steps_per_epoch=conf['session'][session_type]['steps_per_epoch'],
            batch_size=conf['session'][session_type]['batch_size'],
        )

    else:
        logger.error('Invalid choice.')


if __name__ == '__main__':
    args = parse_args()
    np.random.seed(1000)
    run(
        config_file=args['config_file'],
        default=args['default'],
        training=args['training'],
        prediction=args['prediction'],
    )



from setuptools import setup, find_packages

long_description = '''
ISR (Image Super-Resolution) is a library to upscale and improve the quality of low resolution images.

Read the documentation at: https://idealo.github.io/image-super-resolution/

ISR is compatible with Python 3.6 and is distributed under the Apache 2.0 license.
'''

setup(
    name='ISR',
    version='1.9.3',
    author='Francesco Cardinale',
    author_email='testadicardi@gmail.com',
    description='Image Super Resolution',
    long_description=long_description,
    license='Apache 2.0',
    install_requires=['imageio', 'Keras==2.2.4', 'numpy==1.16.2', 'tensorflow==1.13.1', 'tqdm'],
    extras_require={
        'tests': ['pytest==4.3.0', 'pytest-cov==2.6.1'],
        'docs': ['mkdocs==1.0.4', 'mkdocs-material==4.0.2'],
    },
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Intended Audience :: Education',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Topic :: Software Development :: Libraries',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
    packages=find_packages(exclude=('tests',)),
)
